from django.contrib.auth.models import Permission
from django.contrib.auth.models import User
from django.test import TestCase

from django_delayed_union import DelayedDifferenceQuerySet
from django_delayed_union import DelayedIntersectionQuerySet
from django_delayed_union import DelayedUnionQuerySet

from .factories import UserFactory
from .mixins import DelayedQuerySetMetaTestsMixin
from .mixins import DelayedQuerySetTestsMixin


class DelayedUnionQuerySetMetaTests(DelayedQuerySetMetaTestsMixin, TestCase):
    def get_class(self):
        return DelayedUnionQuerySet

    def test_delayed_union_only_accepts_all_as_kwarg(self):
        with self.assertRaises(TypeError):
            DelayedUnionQuerySet(
                User.objects.all(),
                User.objects.all(),
                foo=42
            )

    def test_does_not_support_nested_delayed_querysets_with_incompatible_kwargs(self):
        with self.assertRaises(ValueError):
            DelayedUnionQuerySet(
                User.objects.all(),
                DelayedUnionQuerySet(
                    User.objects.all(),
                    User.objects.all(),
                    all=True
                )
            )

    def test_does_not_support_delayed_intersections_as_querysets(self):
        with self.assertRaises(ValueError):
            DelayedUnionQuerySet(
                User.objects.all(),
                DelayedIntersectionQuerySet(
                    User.objects.all(),
                    User.objects.all(),
                )
            )

    def test_does_not_support_delayed_differences_as_querysets(self):
        with self.assertRaises(ValueError):
            DelayedUnionQuerySet(
                User.objects.all(),
                DelayedDifferenceQuerySet(
                    User.objects.all(),
                    User.objects.all(),
                )
            )


class DelayedUnionQuerySetTestsMixin(DelayedQuerySetTestsMixin):
    def test_select_related(self):
        base_qs = Permission.objects.all()
        qs = DelayedUnionQuerySet(base_qs, base_qs)

        self.assertTrue(qs.exists())
        for permission in qs.select_related('content_type'):
            with self.assertNumQueries(0):
                self.assertIsNotNone(permission.content_type.id)

    def test_update(self):
        self.qs.update(first_name='Rover')
        for user in self.qs:
            self.assertEqual(user.first_name, 'Rover')


class DelayedUnionQuerySetTests(DelayedUnionQuerySetTestsMixin, TestCase):
    def get_queryset(self):
        return DelayedUnionQuerySet(
            User.objects.all(),
            User.objects.all(),
        )

    def get_expected_models(self):
        return [self.user]


class DelayedUnionQuerySetMixedTests(
        DelayedUnionQuerySetTestsMixin,
        TestCase):

    def get_queryset(self):
        return DelayedUnionQuerySet(
            User.objects.filter(id=self.bad_id),
            User.objects.all(),
        )

    def get_expected_models(self):
        return [self.user]


class DelayedUnionQuerySetReversedMixedTests(
        DelayedUnionQuerySetTestsMixin,
        TestCase):

    def get_queryset(self):
        return DelayedUnionQuerySet(
            User.objects.all(),
            User.objects.filter(id=self.bad_id),
        )

    def get_expected_models(self):
        return [self.user]


class DelayedUnionAllMutuallyExclusiveQuerySetTests(
        DelayedUnionQuerySetTestsMixin,
        TestCase):

    def get_queryset(self):
        return DelayedUnionQuerySet(
            User.objects.filter(id=self.user.id),
            User.objects.exclude(id=self.user.id),
            all=True
        )

    def get_expected_models(self):
        return [self.user]


class NestedDelayedUnionQuerySetTests(
        DelayedUnionQuerySetTestsMixin,
        TestCase):

    @classmethod
    def setUpTestData(cls):
        super(NestedDelayedUnionQuerySetTests, cls).setUpTestData()
        cls.user_b, cls.user_c = UserFactory.create_batch(2)

    def get_queryset(self):
        return DelayedUnionQuerySet(
            User.objects.exclude(id=self.user.id),
            DelayedUnionQuerySet(
                User.objects.exclude(id=self.user_b.id),
                User.objects.exclude(id=self.user_c.id),
            )
        )

    def get_expected_models(self):
        return [self.user, self.user_b, self.user_c]


class DelayedUnionAllQuerySetTests(DelayedUnionQuerySetTestsMixin, TestCase):
    @classmethod
    def setUpTestData(cls):
        super(DelayedUnionQuerySetTestsMixin, cls).setUpTestData()
        cls.user_b = UserFactory.create()

    def get_queryset(self):
        return DelayedUnionQuerySet(
            User.objects.all(),
            User.objects.filter(id=self.user_b.id),
            all=True
        )

    def get_expected_models(self):
        return [self.user, self.user_b, self.user_b]

    def test_get_with_duplicates(self):
        with self.assertRaises(User.MultipleObjectsReturned):
            self.qs.get(id=self.user_b.id)



import factory
from django.contrib.auth.models import User
from django.utils import timezone


class UserFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = User

    username = factory.Sequence(lambda n: 'username{}'.format(n))
    email = factory.Sequence(lambda n: 'email{}@rover.com'.format(n))

    first_name = factory.Sequence(lambda n: '\xd3scarNumber{}'.format(n))
    last_name = factory.Sequence(lambda n: 'Ib\xe1\xf1ezNumber{}'.format(n))

    date_joined = factory.LazyAttribute(lambda user: timezone.now())

    @classmethod
    def _setup_next_sequence(cls):
        return 0



from django.contrib.auth.models import User
from django.db import models


class UserProfile(models.Model):
    user = models.OneToOneField(
        User,
        related_name='user_profile',
        on_delete=models.PROTECT
    )



from django_delayed_union.base import DelayedQuerySetDescriptor
from sphinx.ext.autodoc import AttributeDocumenter
from sphinx.ext.autodoc import DocstringSignatureMixin


class DelayedQuerySetDescriptorDocumenter(AttributeDocumenter):
    objtype = 'delayedquerysetdescriptor'
    directivetype = 'attribute'
    priority = AttributeDocumenter.priority + 1

    @classmethod
    def can_document_member(cls, member, membername, isattr, parent):
        return isinstance(member, DelayedQuerySetDescriptor)

    def format_signature(self):
        return DocstringSignatureMixin.format_signature(self)


def setup(app):
    app.add_autodocumenter(DelayedQuerySetDescriptorDocumenter)
    return {'version': '0.1'}   # identifies the version of our extension



from django.contrib.auth.models import Permission
from django.contrib.auth.models import User
from django.test import TestCase

from django_delayed_union import DelayedIntersectionQuerySet

from .factories import UserFactory
from .markers import skip_for_mysql
from .mixins import DelayedQuerySetMetaTestsMixin
from .mixins import DelayedQuerySetTestsMixin


class DelayedIntersectionQuerySetMetaTests(
        DelayedQuerySetMetaTestsMixin,
        TestCase):

    def get_class(self):
        return DelayedIntersectionQuerySet


@skip_for_mysql
class DelayedIntersectionQuerySetTestsMixin(DelayedQuerySetTestsMixin):
    def test_select_related(self):
        base_qs = Permission.objects.all()
        qs = DelayedIntersectionQuerySet(base_qs, base_qs)

        self.assertTrue(qs.exists())
        for permission in qs.select_related('content_type'):
            with self.assertNumQueries(0):
                self.assertIsNotNone(permission.content_type.id)


@skip_for_mysql
class DelayedIntersectionQuerySetTests(
        DelayedIntersectionQuerySetTestsMixin,
        TestCase):

    @classmethod
    def setUpTestData(cls):
        super(DelayedIntersectionQuerySetTests, cls).setUpTestData()
        cls.user_b, cls.user_c = UserFactory.create_batch(2)

    def get_queryset(self):
        return DelayedIntersectionQuerySet(
            User.objects.exclude(id=self.user_b.id),
            User.objects.exclude(id=self.user_c.id)
        )

    def get_expected_models(self):
        return [self.user]


@skip_for_mysql
class NestedDelayedIntersectionQuerySetTests(
        DelayedIntersectionQuerySetTestsMixin,
        TestCase):

    @classmethod
    def setUpTestData(cls):
        super(NestedDelayedIntersectionQuerySetTests, cls).setUpTestData()
        cls.user_b, cls.user_c, cls.user_d = UserFactory.create_batch(3)

    def get_queryset(self):
        return DelayedIntersectionQuerySet(
            User.objects.exclude(id=self.user_b.id),
            DelayedIntersectionQuerySet(
                User.objects.exclude(id=self.user_c.id),
                User.objects.exclude(id=self.user_d.id),
            )
        )

    def get_expected_models(self):
        return [self.user]



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import io
import re
from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import splitext

from setuptools import find_packages
from setuptools import setup


def read(*names, **kwargs):
    return io.open(
        join(dirname(__file__), *names),
        encoding=kwargs.get('encoding', 'utf8')
    ).read()


setup(
    name='django-delayed-union',
    version='0.1.6',
    license='BSD 3-Clause License',
    description="A library designed to workaround some drawbacks with Django's union, intersection, and difference operations.",
    long_description='%s\n%s' % (
        re.compile('^.. start-badges.*^.. end-badges', re.M | re.S).sub('', read('README.rst')),
        re.sub(':[a-z]+:`~?(.*?)`', r'``\1``', read('CHANGELOG.rst'))
    ),
    author='Mike Hansen',
    author_email='mike@rover.com',
    url='https://github.com/roverdotcom/django-delayed-union',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Framework :: Django',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 2.1',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.0',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Utilities',
    ],
    keywords=[
        # eg: 'keyword1', 'keyword2', 'keyword3',
    ],
    install_requires=[
        'django>=2.2',
    ],
    tests_require=[
        'factory_boy'
    ],
    extras_require={
        # eg:
        #   'rst': ['docutils>=0.11'],
        #   ':python_version=="2.6"': ['argparse'],
    },
)
